@@ -1,12 +1,13 @@
-use anyhow::{anyhow, Result};
-use argh::from_env;
-use cli::{Interface, SubCommandEnum};
+use anyhow::{anyhow, bail, Result};
+use clap::Parser;
+use cli::{Opts, SubCommands};
 use core::default::Default;
 use dialoguer::{theme::ColorfulTheme, Select};
+use regex::Regex;
 use serde::{Deserialize, Serialize};
 use std::{
     fs,
-    path::{Path, PathBuf},
+    path::PathBuf,
     process::{Command, Stdio},
 };
 
@@ -17,10 +18,10 @@ mod cli;
 mod i18n;
 mod kernel;
 mod macros;
+mod parser;
 
 const CONF_PATH: &str = "/etc/systemd-boot-friend.conf";
 const REL_DEST_PATH: &str = "EFI/systemd-boot-friend/";
-const INSTALLED_PATH: &str = "/var/lib/systemd-boot-friend/installed.json";
 
 #[derive(Debug, Serialize, Deserialize)]
 pub struct Config {
@@ -39,8 +40,8 @@ pub struct Config {
 impl Default for Config {
     fn default() -> Self {
         Config {
-            vmlinuz: "vmlinuz-{VERSION}-{LOCALVERSION}".to_owned(),
-            initrd: "initramfs-{VERSION}-{LOCALVERSION}.img".to_owned(),
+            vmlinuz: "vmlinuz-{VERSION}".to_owned(),
+            initrd: "initramfs-{VERSION}.img".to_owned(),
             distro: "AOSC OS".to_owned(),
             esp_mountpoint: PathBuf::from("/efi"),
             bootarg: String::new(),
@@ -48,23 +49,35 @@ impl Default for Config {
     }
 }
 
-/// Read config, create a default one if the file is missing
-fn read_config() -> Result<Config> {
-    fs::read(CONF_PATH).map_or_else(
-        |_| {
-            println_with_prefix_and_fl!("conf_default", conf_path = CONF_PATH);
-            fs::create_dir_all(PathBuf::from(CONF_PATH).parent().unwrap())?;
-            fs::write(CONF_PATH, toml::to_string_pretty(&Config::default())?)?;
-            Err(anyhow!(fl!("edit_conf", conf_path = CONF_PATH)))
-        },
-        |f| Ok(toml::from_slice(&f)?),
-    )
+/// Generate installed kernel list
+fn list_installed_kernels(config: &Config) -> Result<Vec<Kernel>> {
+    // Construct regex for the template
+    let re = Regex::new(&config.vmlinuz.replace("{VERSION}", r"(?P<version>.+)"))?;
+    // Regex match group
+    let mut installed_kernels = Vec::new();
+    if let Ok(d) = fs::read_dir(config.esp_mountpoint.join(REL_DEST_PATH)) {
+        for x in d {
+            let filename = &x?
+                .file_name()
+                .into_string()
+                .map_err(|_| anyhow!(fl!("invalid_kernel_filename")))?;
+            if let Some(c) = re.captures(filename) {
+                let version = c
+                    .name("version")
+                    .ok_or_else(|| anyhow!(fl!("invalid_kernel_filename")))?
+                    .as_str();
+                installed_kernels.push(Kernel::parse(config, version)?);
+            }
+        }
+    }
+
+    Ok(installed_kernels)
 }
 
 /// Choose a kernel using dialoguer
 fn choose_kernel(kernels: &[Kernel]) -> Result<Kernel> {
     if kernels.is_empty() {
-        return Err(anyhow!(fl!("empty_list")));
+        bail!(fl!("empty_list"));
     }
     // build dialoguer Select for kernel selection
     let n = Select::with_theme(&ColorfulTheme::default())
@@ -76,20 +89,19 @@ fn choose_kernel(kernels: &[Kernel]) -> Result<Kernel> {
 }
 
 /// Update systemd-boot kernels and entries
-fn update(config: &Config, installed_kernels: &mut Vec<Kernel>, kernels: &[Kernel]) -> Result<()> {
-    while let Some(k) = installed_kernels.pop() {
-        k.remove(config)?;
+fn update(installed_kernels: &[Kernel], kernels: &[Kernel]) -> Result<()> {
+    for k in installed_kernels.iter() {
+        k.remove()?;
     }
     for k in kernels.iter() {
-        k.install_and_make_config(config, true)?;
-        installed_kernels.push(k.clone());
+        k.install_and_make_config(true)?;
     }
 
     Ok(())
 }
 
 /// Initialize the default environment for friend
-fn init(config: &Config, installed_kernels: &mut Vec<Kernel>, kernels: &[Kernel]) -> Result<()> {
+fn init(config: &Config, installed_kernels: &[Kernel], kernels: &[Kernel]) -> Result<()> {
     // use bootctl to install systemd-boot
     println_with_prefix_and_fl!("initialize");
     Command::new("bootctl")
@@ -107,7 +119,7 @@ fn init(config: &Config, installed_kernels: &mut Vec<Kernel>, kernels: &[Kernel]
     println_with_prefix_and_fl!("create_folder");
     fs::create_dir_all(config.esp_mountpoint.join(REL_DEST_PATH))?;
     // Update systemd-boot kernels and entries
-    update(config, installed_kernels, kernels)
+    update(installed_kernels, kernels)
 }
 
 #[inline]
@@ -117,63 +129,73 @@ fn parse_num_or_filename(config: &Config, n: &str, kernels: &[Kernel]) -> Result
             .get(num - 1)
             .ok_or_else(|| anyhow!(fl!("invalid_index")))?
             .clone(),
-        Err(_) => Kernel::parse(n, config)?,
+        Err(_) => Kernel::parse(config, n)?,
     })
 }
 
+fn read_config() -> Result<Config> {
+    fs::read(CONF_PATH).map_or_else(
+        |_| {
+            println_with_prefix_and_fl!("conf_default", conf_path = CONF_PATH);
+            fs::create_dir_all(PathBuf::from(CONF_PATH).parent().unwrap())?;
+            fs::write(CONF_PATH, toml::to_string_pretty(&Config::default())?)?;
+            bail!(fl!("edit_conf", conf_path = CONF_PATH))
+        },
+        |f| {
+            let mut config: Config = toml::from_slice(&f)?;
+            // Migrate from old configuration
+            let old_conf = "{VERSION}-{LOCALVERSION}";
+            let new_conf = "{VERSION}";
+            if config.vmlinuz.contains(old_conf)
+                || config.initrd.contains(old_conf)
+            {
+                println_with_prefix_and_fl!("conf_old");
+                config.vmlinuz = config.vmlinuz.replace(old_conf, new_conf);
+                config.initrd = config.initrd.replace(old_conf, new_conf);
+                fs::write(CONF_PATH, toml::to_string_pretty(&config)?)?;
+            }
+            Ok(config)
+        },
+    )
+}
+
 fn main() -> Result<()> {
     // CLI
-    let matches: Interface = from_env();
-    if matches.version {
-        println_with_prefix!(env!("CARGO_PKG_VERSION"));
-        return Ok(());
-    }
-    // Read config
+    let matches: Opts = Opts::parse();
+    // Read config, create a default one if the file is missing
     let config = read_config()?;
-    // the record file of installed kernels, use empty value if not found
-    let mut installed_kernels = Vec::new();
-    if let Ok(f) = fs::read(INSTALLED_PATH) {
-        installed_kernels = serde_json::from_slice::<Vec<String>>(&f)?
-            .iter()
-            .map(|s| Kernel::parse(s, &config))
-            .collect::<Result<Vec<Kernel>>>()?;
-    } else {
-        // Create the folder structure for the record of installed kernels
-        fs::create_dir_all(Path::new(INSTALLED_PATH).parent().unwrap())?;
-        serde_json::to_writer(fs::File::create(INSTALLED_PATH)?, &Vec::<String>::new())?;
-    }
+    let installed_kernels = list_installed_kernels(&config)?;
     let kernels = Kernel::list_kernels(&config)?;
     // Switch table
-    match matches.nested {
+    match matches.subcommands {
         Some(s) => match s {
-            SubCommandEnum::Init(_) => init(&config, &mut installed_kernels, &kernels)?,
-            SubCommandEnum::List(_) => {
+            SubCommands::Init(_) => init(&config, &installed_kernels, &kernels)?,
+            SubCommands::List(_) => {
                 // list available kernels
-                for (i, k) in Kernel::list_kernels(&config)?.iter().enumerate() {
+                for (i, k) in kernels.iter().enumerate() {
                     println!("\u{001b}[1m[{}]\u{001b}[0m {}", i + 1, k);
                 }
             }
-            SubCommandEnum::Install(args) => {
+            SubCommands::Install(args) => {
                 let kernel = match args.target {
                     // the target can be both the number in
                     // the list and the name of the kernel
                     Some(n) => parse_num_or_filename(&config, &n, &kernels)?,
                     // install the newest kernel
                     // when no target is given
-                    None => Kernel::list_kernels(&config)?
+                    None => kernels
                         .first()
                         .ok_or_else(|| anyhow!(fl!("no_kernel")))?
                         .clone(),
                 };
-                kernel.install_and_make_config(&config, args.force)?;
-                installed_kernels.push(kernel);
+                kernel.install_and_make_config(args.force)?;
             }
-            SubCommandEnum::ListInstalled(_) => {
+            SubCommands::ListInstalled(_) => {
                 for (i, k) in installed_kernels.iter().enumerate() {
                     println!("\u{001b}[1m[{}]\u{001b}[0m {}", i + 1, k);
                 }
             }
-            SubCommandEnum::Remove(args) => {
+            SubCommands::Remove(args) => {
                 let kernel = match args.target {
                     // the target can be both the number in
                     // the list and the name of the kernel
@@ -182,28 +204,12 @@ fn main() -> Result<()> {
                     // when no target is given
                     None => choose_kernel(&installed_kernels)?,
                 };
-                kernel.remove(&config)?;
-                installed_kernels.retain(|k| *k != kernel);
+                kernel.remove()?;
             }
-            SubCommandEnum::Update(_) => update(
-                &config,
-                &mut installed_kernels,
-                &Kernel::list_kernels(&config)?,
-            )?,
+            SubCommands::Update(_) => update(&installed_kernels, &kernels)?,
         },
-        None => {
-            let kernel = choose_kernel(&Kernel::list_kernels(&config)?)?;
-            // make sure the kernel selected is installed
-            kernel.install_and_make_config(&config, false)?;
-            installed_kernels.push(kernel);
-        }
+        None => choose_kernel(&kernels)?.install_and_make_config(false)?,
     }
-    // Write the installed kernels file
-    installed_kernels.sort();
-    installed_kernels.dedup();
-    let installed_kernels: Vec<String> =
-        installed_kernels.iter().map(|k| format!("{}", k)).collect();
-    serde_json::to_writer(fs::File::create(INSTALLED_PATH)?, &installed_kernels)?;
 
     Ok(())
 }
